/* --------------------------------------------------------------- 
  Dom::UnivariateRationalFunctions(x, F) -- Domain to represent the rational
                                  functions in x over the field F
------------------------------------------------------------------ 

* Call: Dom::UnivariateRationalFunctions(x, F)

* Parameters: x -- an indeterminate
              F -- a domain of Cat::Field

* could, of course, be done by 
  Dom::Fraction(Dom::UnivariatePolynomial(...))
  but this is more efficient and more flexible.

* created mainly for use in Dom::ShiftOperators
              
* ... blah ...

* Author: Peter Horn, hornp@xinit.org, June 2004
--------------------------------------------------------------- */

domain Dom::UnivariateRationalFunctions(Indet=x, F=Dom::Rational)
  local X, R, EM;
  inherits Dom::BaseDomain;
  category Cat::Field; //, Cat::Algebra(F), Cat::VectorSpace(F);

  //axiom if F::hasProp(Ax::canonicalRep) then Ax::canonicalRep else null() end,
  //      if F::hasProp(Ax::normalRep) then Ax::normalRep else null() end;
             //, Cat::Algebra(Field), Cat::VectorSpace(Field);

// Basic Properties of the Function-Field
BaseField := F;
Indet := X;
//characteristic := F::characteristic;
             
//------------------------------------------------------------------
// new
//------------------------------------------------------------------
new := proc(e) local n,d;
  begin
    if args(0) <> 1 then error("Expecting ONE argument"); end_if;
  if domtype(e) = dom then return(e) 
  elif domtype(e) = DOM_LIST then 
      if nops(e) <> 2 then error("Wrong kind of arguments"); end_if;
    n := R(e[1]);
    d := R(e[2]);
  else
    if hastype(e,DOM_POLY) then e := expr(e) end_if;
    traperror((e := normal(e)));
    n := R(e);
    if n <> FAIL then
      return (new(dom, n, R(1)));
    end_if;
    n := R(numer(e));
    d := R(denom(e));
  end_if;
  if FAIL in {n, d} then return(FAIL); end_if;
    [n,d] := dom::dnormal(n,d);
  new(dom, n, d);
  end; // new

convert := dom::new;

//------------------------------------------------------------------
// dnormal -- internally used to normalize
//------------------------------------------------------------------
dnormal := proc(n : DOM_POLY, d : DOM_POLY) local g, l; 
  begin
    g := gcd(n, d);
    if expr(g) <> 1 then
      n := divide(n,g)[1];
      d := divide(d,g)[1];
    end_if;
    l := lcoeff(d);
    if expr(l) <> 1 then
      n := multcoeffs(n, 1/l);
      d := multcoeffs(d, 1/l);
    end_if;
    if F = Dom::ExpressionField() then
      n := mapcoeffs(n, normal);
      d := mapcoeffs(d, normal);
    end_if;
    [n, d];
  end; // dnormal
    
//------------------------------------------------------------------
// normal, numer, denom
//------------------------------------------------------------------
normal := proc(p:dom) begin p end_proc;
numer := proc(r:dom) begin extop(r,1); end_proc;
denom := proc(r:dom) begin extop(r,2); end_proc;

//------------------------------------------------------------------
// Constants zero, one
//------------------------------------------------------------------
zero := dom::new(0);
one := dom::new(1);

//------------------------------------------------------------------
// _plus
//------------------------------------------------------------------
_plus := proc()
  local A, a, t, f1, f2;
  begin
    A := select([args()], (a)->not(iszero(a)));
    if nops(A) = 0 then return(dom::zero) end_if;
    A := [ _if(domtype(a) = dom, a, _if((t := dom::new(a)) <> FAIL, t, a)) $ a in A];
    A := split(A, (a) -> domtype(a)=dom);
    f1 := dom::_plus2(op(A[1]));
    if (nops(A[2])+nops(A[3])>0) then
      f2 := _plus(op(A[2].A[3]));
      return(hold(_plus)(f1,f2));
    else
      return(f1);
    end_if;
  end_proc; // _plus

//------------------------------------------------------------------
// _plus2 -- does the actual addition
//------------------------------------------------------------------
_plus2 := proc(r, s) local rn, rd, i, n1, d1, n2, d2;
  begin
  if args(0) = 0 then return(dom::zero); end_if;
  if args(0) = 1 then return(r); end_if;
  if {domtype(r), domtype(s)} <> {dom} then
    error("incompatible type for add -- this shouldn't happen, though!");
  end_if;
  n1 := dom::numer(r);
  d1 := dom::denom(r);
  n2 := dom::numer(s);
  d2 := dom::denom(s);
  if d1 = d2 then
    rn := n1+n2;
    rd := d1;
  else
    rn := n1*d2 + n2*d1;
    rd := d1*d2;
    [rn, rd] := dom::dnormal(rn, rd);
  end_if;
  if iszero(rn) then return(dom::zero) end_if;
  if args(0) = 2 then return(new(dom, rn, rd)); end_if;
  return (dom::_plus2(new(dom, rn, rd), args(i) $ i=3..args(0)));  
  end_proc; // _plus2

//------------------------------------------------------------------
// _substract
//------------------------------------------------------------------
_substract := proc(r, s)
  begin
    dom::_plus(r, _negate(s));
  end; // _substract

//------------------------------------------------------------------
// _negate
//------------------------------------------------------------------
_negate := proc(r) begin
    if args(0) <> 1 then error("One argument expected"): end_if;
    if domtype(r) <> dom then
      dom::new(-1*r);
    else
      new(dom, multcoeffs(dom::numer(r), -1), dom::denom(r));
    end_if;
  end_proc; // _negate
  
//------------------------------------------------------------------
// _mult
//------------------------------------------------------------------
_mult := proc()
  local A, a, t, f1, f2;
  begin
    A := [ _if(domtype(a) = dom, a, _if((t := dom::new(a)) <> FAIL, t, a)) $ a in args()];
    A := split(A, (a) -> domtype(a)=dom);
    f1 := dom::_mult2(op(A[1]));
    if (nops(A[2])+nops(A[3])>0) then
      if iszero(f1) then
        return(0)
      else
        f2 := _mult(op(A[2].A[3]));
        return(hold(_mult)(f1,f2));
      end_if;
    else
      return(f1);
    end_if;
  end_proc; // _mult
  
//------------------------------------------------------------------
// _mult2 -- does the actual multiplication
//------------------------------------------------------------------
_mult2 := proc(r, s) local rn, rd, i, d1, d2, n1, n2, g;
  begin
    if args(0) = 0 then return(dom::one); end_if;
    if args(0) = 1 then return(r); end_if;
    if {domtype(r), domtype(s)} <> {dom} then
      error("incompatible type for multiply -- this shouldn't happen, though!");
    end_if;    
    n1 := dom::numer(r);
    d1 := dom::denom(r);
    n2 := dom::numer(s);
    d2 := dom::denom(s);
    g := gcd(dom::denom(r), dom::numer(s)) * gcd(dom::denom(s), dom::numer(r));
    
    rn := divide(n1*n2, g, Quo);
    rd := divide(d1*d2, g, Quo);
    
    if degree(rd,X) = 0 then
      // rare case if ground-field is Dom::ExpressionField()
      rn := multcoeffs(rn, 1/expr(rd));
      rd := R(1);
    end_if;
    
    //if rd <> dom::one then
    //  [rn, rd] := dom::dnormal(rn, rd);
    //end_if;
    if args(0) = 2 then return(new(dom, rn, rd)); end_if;
    return (dom::_mult2(new(dom, rn, rd), args(i) $ i=3..args(0)));  
  end_proc; // _mult2

//------------------------------------------------------------------
// _divide
//------------------------------------------------------------------
_divide := proc(r, s) local rn, rd;
  begin
    if args(0) <> 2 then error("Two parameters expected"); end_if;
    if domtype(r) <> dom then r := dom::new(r) end_if;
    if domtype(s) <> dom then s := dom::new(s) end_if;
    rn := dom::numer(r)*dom::denom(s);
    rd := dom::denom(r)*dom::numer(s);
    [rn, rd] := dom::dnormal(rn, rd);
    return(new(dom, rn, rd)); 
  end; // _divide

//------------------------------------------------------------------
// _invert
//------------------------------------------------------------------
_invert := proc(r : dom) 
  begin
    if args(0) <> 1 then error("One argument expected"): end_if;
    if domtype(r) <> dom then r := dom::new(r) end_if;
    if r = dom::zero then error("Division by zero"): end_if;
    new(dom, op(dom::dnormal(dom::denom(r), dom::numer(r))));
  end_proc; // _invert
  
//------------------------------------------------------------------
// _power
//------------------------------------------------------------------ 
_power := proc(r : dom, e : DOM_INT) 
  local rn, rd;
  begin
    rn := dom::numer(r)^abs(e);
    rd := dom::denom(r)^abs(e);
    if e<0 then
      [rn, rd] := [rd, rn];          
      e := -e;
      [rn, rd] := dom::dnormal(rn, rd);
    end_if;
    return(new(dom, rn, rd)); 
  end_proc; // _power

//------------------------------------------------------------------
// _mod
//------------------------------------------------------------------ 
modp := dom::_mod;
_mod := proc(r : dom, e : DOM_INT) 
  local rn, rd;
  begin
    rn := mapcoeffs(dom::numer(r), _mod, e);
    rd := mapcoeffs(dom::denom(r), _mod, e);
    if iszero(rd) then
      error("Division by zero");
    end_if;
    if iszero(rn) then
      return (dom::zero);
    end_if;
    return(new(dom, rn, rd)); 
  end_proc; // _power

//------------------------------------------------------------------
// iszero
//------------------------------------------------------------------
iszero := (r) -> iszero(dom::numer(r));

//------------------------------------------------------------------
// func_call -- evaluates the dom at a point
//------------------------------------------------------------------
func_call := proc(r:dom, a) 
  begin
    r := context(r);
    a := context(a);
    expr(dom::numer(r))/expr(dom::denom(r)) | X = a;
  end_proc; // func_call

//------------------------------------------------------------------
// shift(f, a) -- computes the shifted dom f(x+a)
//------------------------------------------------------------------
shift := proc(r:dom, a) 
  begin
    r := context(r);
    a := context(a);
    dom::new([ polylib::shift(dom::numer(r), a), polylib::shift(dom::denom(r), a)  ])
  end_proc; // shift

//------------------------------------------------------------------
// qShift -- computes the q-shifted dom
//            of doms
//------------------------------------------------------------------
qShift := proc(r, q, k) 
  local n,d;
  begin
    n := poly2list(dom::numer(r));
    d := poly2list(dom::denom(r)); 
    n := R(map(n, (x) -> [q^(k*x[2])*x[1],x[2]]));
    d := R(map(d, (x) -> [q^(k*x[2])*x[1],x[2]]));
    if n=FAIL or d=FAIL then
      error("damn!")
    end_if;
    return(new(dom, n, d));
  end;

//------------------------------------------------------------------
// unitNormal -- 1 if r is <> 0, 0 otherwise
//------------------------------------------------------------------
unitNormal := proc(r:dom) 
  begin
    _if(r=dom::zero, dom::zero, dom::one);
  end_proc; // unitNormal
  
//------------------------------------------------------------------
// polyprint -- a nice way to represent polys
//------------------------------------------------------------------
polyprint := proc(p : DOM_POLY) 
local i, l, St;
begin
//  return(op(rep(p),1)); 
  if domtype(p) <> DOM_POLY then 
    error("POLYBRAUCHTS! - ".expr2text(domtype(p))." - ".expr2text(hold(p))); 
  end_if;
  
  St := (x) ->
      _if(x[2]=0,
        x[1],
        _if(x[1]=1,
          X^x[2],
          hold(_mult)(x[1],X^x[2])));

  l := poly2list(p);
  if nops(l) = 0 then
    0
  elif nops(l)<=1 then
    St(l[1]);
  else
    hold(_plus)(St(i)$ i in l);
  end_if;
end_proc; // polyprint

//------------------------------------------------------------------
// print -- putput
//------------------------------------------------------------------
print := proc(r) 
  begin 
    if degree(dom::denom(r)) = 0 then
      dom::polyprint(dom::numer(r));
    else
      hold(_divide)(dom::polyprint(dom::numer(r)), dom::polyprint(dom::denom(r)));
    end_if; 
  end_proc; // print
/*  begin 
    if degree(dom::denom(r)) = 0 then
      "(".expr2text(dom::polyprint(dom::numer(r))).")";
    else
      "((".expr2text(dom::polyprint(dom::numer(r))).")/("
        .expr2text(dom::polyprint(dom::denom(r)))."))"; 
    end_if; 
  end_proc; // print
*/
//------------------------------------------------------------------
// expr -- return the corresponding expression
//------------------------------------------------------------------
expr := proc(r) 
  begin 
    if degree(dom::denom(r)) = 0 then
      (expr(dom::numer(r)));
    else
      (expr(dom::numer(r))/expr(dom::denom(r))); 
    end_if;  
  end_proc; // expr
  
//------------------------------------------------------------------
// D -- differentiate wrt. the given indeterminate
//------------------------------------------------------------------
D := proc(r : dom) local g,f,df,dg;
  begin
    f := dom::numer(r); df := diff(f,X);
    g := dom::denom(r); dg := diff(g,X);
    dom::new([(df*g-f*dg),(g*g)]);
  end_proc; // D

//------------------------------------------------------------------
// lcdenom -- computes the least common denominator of a sequence 
//            of doms
//------------------------------------------------------------------
lcdenom := proc() 
  local A, a, i , t;
  begin
    A := [ _if(domtype(a) = dom, 
              a, 
              _if((t := dom::new(a)) <> FAIL, t, a)) $ a in args()];
    A := split(A, (a) -> domtype(a)=dom);
    for i in [2,3] do
      if nops(A[i]) > 0 then
        A[i] := map(A[i], dom::new);
        if has(A[i], FAIL) then return(FAIL) end_if;
      end_if; 
    end_for;
    return(lcm(op(map(A[1].A[2].A[3], denom))));
  end_proc; // lcdenom
  
//==================================================================
//* Domain Constructor
//------------------------------------------------------------------
begin
  EM := "Call: RationalFunctions(F, I) where F is a field and I is an ".
        "indeterminate";
  //if args(0) <> 2 then error("Wrong number of Arguments - \n".EM); end_if;
/*  if not (F::hasProp(Cat::Field) and domtype(Indet) = DOM_IDENT) then
    error("Wrong kind of Arguments - \n".EM);
  end_if;
*/  X := Indet;
  if F=Dom::ExpressionField() then
    R := () -> poly(args(1), [X], Expr);
  elif op(F,0)=IntMod then
    R := () -> poly(args(1), [X], F);
  else
    R := () -> poly(args(1), [X], F);
  end_if; 
/*  if F in { Dom::ExpressionField(), Expr } then
    R := proc(p) local i; begin;
      i := indets(p) minus {X};
      poly(p, [X], F);
      if nops(i) > 0 then
        mapcoeffs(%, (x)->collect(x, [op(i)]));
      end_if;
    end_proc;
  else
    R := proc(p) local i; begin;
      i := indets(p) minus {X};
      poly(p, [X], Expr);
      if nops(i) > 0 then
        mapcoeffs(%, (x)->collect(x, [op(i)]));
      end_if;
    end_proc;
  end_if;
  */
end:

//==================================================================
// End of file RatFun.mu
//==================================================================

